import logging
import re

from os import environ
from urllib.parse import urlparse

import elasticsearch
import elasticsearch.helpers

from assemblyline.common import forge
from assemblyline.datastore.collection import ESCollection
from assemblyline.datastore.exceptions import DataStoreException

TRANSPORT_TIMEOUT = int(environ.get('AL_DATASTORE_TRANSPORT_TIMEOUT', '10'))


class ESStore(object):
    """ Elasticsearch multi-index implementation of the ResultStore interface."""
    DEFAULT_SORT = "id asc"
    DATE_FORMAT = {
        'NOW': 'now',
        'YEAR': 'y',
        'MONTH': 'M',
        'WEEK': 'w',
        'DAY': 'd',
        'HOUR': 'h',
        'MINUTE': 'm',
        'SECOND': 's',
        'MILLISECOND': 'ms',
        'MICROSECOND': 'micros',
        'NANOSECOND': 'nanos',
        'SEPARATOR': '||',
        'DATE_END': 'Z'
    }
    DATEMATH_MAP = {
        'NOW': 'now',
        'YEAR': 'y',
        'MONTH': 'M',
        'WEEK': 'w',
        'DAY': 'd',
        'HOUR': 'h',
        'MINUTE': 'm',
        'SECOND': 's',
        'DATE_END': 'Z||'
    }
    ID = 'id'

    def __init__(self, hosts, collection_class=ESCollection, archive_access=True):
        config = forge.get_config()
        if config.datastore.ilm.enabled:
            ilm_config = config.datastore.ilm.indexes.as_primitives()
        else:
            ilm_config = {}

        self._hosts = hosts
        self._closed = False
        self._collections = {}
        self._models = {}
        self.ilm_config = ilm_config
        self.validate = True

        tracer = logging.getLogger('elasticsearch')
        tracer.setLevel(logging.CRITICAL)

        self.client = elasticsearch.Elasticsearch(hosts=hosts,
                                                  connection_class=elasticsearch.RequestsHttpConnection,
                                                  max_retries=0,
                                                  timeout=TRANSPORT_TIMEOUT)
        self.archive_access = archive_access
        self.url_path = 'elastic'

    def __enter__(self):
        return self

    def __exit__(self, ex_type, exc_val, exc_tb):
        self.close()

    def __str__(self):
        return '{0} - {1}'.format(self.__class__.__name__, self._hosts)

    def __getattr__(self, name) -> ESCollection:
        if not self.validate:
            return ESCollection(self, name, model_class=self._models[name], validate=self.validate)

        if name not in self._collections:
            self._collections[name] = ESCollection(
                self, name, model_class=self._models[name], validate=self.validate)

        return self._collections[name]

    @property
    def now(self):
        return self.DATE_FORMAT['NOW']

    @property
    def ms(self):
        return self.DATE_FORMAT['MILLISECOND']

    @property
    def us(self):
        return self.DATE_FORMAT['MICROSECOND']

    @property
    def ns(self):
        return self.DATE_FORMAT['NANOSECOND']

    @property
    def year(self):
        return self.DATE_FORMAT['YEAR']

    @property
    def month(self):
        return self.DATE_FORMAT['MONTH']

    @property
    def week(self):
        return self.DATE_FORMAT['WEEK']

    @property
    def day(self):
        return self.DATE_FORMAT['DAY']

    @property
    def hour(self):
        return self.DATE_FORMAT['HOUR']

    @property
    def minute(self):
        return self.DATE_FORMAT['MINUTE']

    @property
    def second(self):
        return self.DATE_FORMAT['SECOND']

    @property
    def date_separator(self):
        return self.DATE_FORMAT['SEPARATOR']

    def connection_reset(self):
        self.client = elasticsearch.Elasticsearch(hosts=self._hosts,
                                                  connection_class=elasticsearch.RequestsHttpConnection,
                                                  max_retries=0,
                                                  timeout=TRANSPORT_TIMEOUT)

    def close(self):
        self._closed = True
        self.client = None

    def get_hosts(self, safe=False):
        if not safe:
            return self._hosts
        else:
            out = []
            for h in self._hosts:
                parsed = urlparse(h)
                out.append(parsed.hostname or parsed.path)
            return out

    def get_models(self):
        return self._models

    def is_closed(self):
        return self._closed

    def ping(self):
        return self.client.ping()

    def register(self, name: str, model_class=None):
        if re.match(r'[a-z0-9_]*', name).string != name:
            raise DataStoreException('Invalid characters in model name. '
                                     'You can only use lower case letters, numbers and underscores.')

        self._models[name] = model_class

    def to_pydatemath(self, value):
        replace_list = [
            (self.now, self.DATEMATH_MAP['NOW']),
            (self.year, self.DATEMATH_MAP['YEAR']),
            (self.month, self.DATEMATH_MAP['MONTH']),
            (self.week, self.DATEMATH_MAP['WEEK']),
            (self.day, self.DATEMATH_MAP['DAY']),
            (self.hour, self.DATEMATH_MAP['HOUR']),
            (self.minute, self.DATEMATH_MAP['MINUTE']),
            (self.second, self.DATEMATH_MAP['SECOND']),
            (self.DATE_FORMAT['DATE_END'], self.DATEMATH_MAP['DATE_END'])
        ]

        for x in replace_list:
            value = value.replace(*x)

        return value



from assemblyline import odm
from assemblyline.common import forge
from assemblyline.odm.models.tagging import Tagging

Classification = forge.get_classification()


# Result Metadata
@odm.model(index=False, store=False)
class ResultOntology(odm.Model):
    # Required metadata
    md5 = odm.MD5()                                                             # MD5 of file
    sha1 = odm.SHA1()                                                           # SHA1 of file
    sha256 = odm.SHA256()                                                       # SHA256 of file
    type = odm.Keyword()                                                        # Type of file as identified by Assemblyline
    size = odm.Integer()                                                        # Size of the file in bytes
    filename = odm.Text()                                                       # Name of the file as submitted
    date = odm.Date()                                                           # Date of analysis
    classification = odm.Keyword(default=Classification.UNRESTRICTED)           # Classification of the service result
    service_name = odm.Keyword()                                                # Service Name
    service_version = odm.Keyword()                                             # Service Version
    service_tool_version = odm.Optional(odm.Keyword(default=''))                # Service Tool Version

    # Optional metadata

    # Who's the parent of this file, if any.
    parent = odm.Optional(odm.SHA256())
    # Used to link to the submission id in Assemblyline
    sid = odm.Optional(odm.Keyword())
    # Which instance of Assemblyline did this come from?
    source_system = odm.Optional(odm.Text())
    # Source as specified by submitter (from metadata)
    original_source = odm.Optional(odm.Text())
    # What classification did the submitter submit under?
    submitted_classification = odm.Keyword(default=Classification.UNRESTRICTED)
    # Who submitted this file?
    submitter = odm.Optional(odm.Keyword())
    # Used to link to knowledge base retaining long-term data
    retention_id = odm.Optional(odm.Keyword())
    # What tags did the service associate to the result
    tags = odm.Optional(odm.Compound(Tagging))
    # What tags are related to certain heuristics raised
    # {
    #   "Bad Things happened": {
    #       "network": {
    #           "static": {
    #               "uri": ["bad.domain", ...]
    #               ...
    #           }
    #           ...
    #       }
    #       ...
    #   }
    # }
    heuristics = odm.Optional(odm.Mapping(odm.Compound(Tagging)))



from assemblyline import odm


# Details about a process
class Process(odm.Model):

    # The GUID associated with the process
    guid = odm.Text()
    # The normalized tag of the object
    tag = odm.Optional(odm.Text())

    # Parent process details
    # The GUID associated with the parent process
    pguid = odm.Optional(odm.Text())
    # The normalized tag of the parent object
    ptag = odm.Optional(odm.Text())
    # The image of the parent process that spawned this process
    pimage = odm.Optional(odm.Text())
    # The command line that the parent process ran
    pcommand_line = odm.Optional(odm.Text())
    # The process ID of the parent process
    ppid = odm.Optional(odm.Integer())

    # The process ID
    pid = odm.Optional(odm.Integer())
    # The image of the process
    image = odm.Text(default="<unknown_image>")
    # The command line that the process ran
    command_line = odm.Optional(odm.Text())
    # The time of creation for the process
    start_time = odm.Date()
    # The time of termination for the process
    end_time = odm.Date()
    # The hash of the tree ID
    tree_id = odm.Optional(odm.Text())
    # Human readable tree ID (concatenation of process names)
    rich_id = odm.Optional(odm.Text())
    # The integrity level of the process
    integrity_level = odm.Optional(odm.Text())
    # The hash of the file run
    image_hash = odm.Optional(odm.Text())
    # The original name of the file
    original_file_name = odm.Optional(odm.Text())
